//-----------------------------------------------------------------------------
// Copyright (c) 2000 Andreas Jnsson
//-----------------------------------------------------------------------------
#include "particleapp.h"
#include <stdio.h>

//-----------------------------------------------------------------------------
// CParticleApp()
//-----------------------------------------------------------------------------
CParticleApp::CParticleApp()
{
	m_szWindowTitle = "Particles";

	m_nNumSystems = 3;
}

//-----------------------------------------------------------------------------
// Initialize()
//-----------------------------------------------------------------------------
HRESULT CParticleApp::Initialize(char *szCmdLine)
{
	// Disable writes to z-buffer so that our transparent particles  
	// can be rendered without depth sorting them first
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, FALSE);

	// Note: Do not disable z-test unless you know that there will not be 
	// any solid objects in the screen, otherwise the particles will be 
	// seen through for example walls
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_ZENABLE, D3DZB_FALSE);

	// Load a texture for the particles
    D3DX_SURFACEFORMAT sf = D3DX_SF_UNKNOWN;
	D3DXCreateTextureFromFile(m_pD3DDev, 
							  NULL,
							  0,			// Let D3DX select width, 
							  0,			// height,
							  &sf,			// and pixelformat
							  NULL,			// We know there is no palette  
											// so we can set this to null
							  &m_pParticleTexture, 
							  0,
							  "particle.bmp",
							  D3DX_FT_LINEAR);

	m_pD3DDev->SetTexture(0, m_pParticleTexture);

	// Setup up the texture engine to multiply the texture  
	// color with the diffuse color of the particle 
	// (This is actually the default mode for Direct3D)
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

	// Turn on mipmapping so that it will be 
	// much faster, even with larger textures
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTFP_POINT); 
	
	// Turn on linear filtering to make it look better
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_LINEAR);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTFG_LINEAR);

	// Disable lighting as the diffuse color of the 
	// particle will be set to black otherwise
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE);

	// We want to blend the particles so that the color is added to the screen
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_ONE);
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_ONE);

	// Create a renderer for our particle systems
	m_Renderer.Initialize(m_pD3DX, 2500);

	// Create the particle systems
	m_ParticleSystems[0].Initialize(0xFFFF6622);
	m_ParticleSystems[0].SetEmitterPos(D3DVECTOR(0,2,0));
	m_ParticleSystems[1].Initialize(0xFF66FF22);
	m_ParticleSystems[1].SetEmitterPos(D3DVECTOR(0,2,0));
	m_ParticleSystems[2].Initialize(0xFF2266FF);
	m_ParticleSystems[2].SetEmitterPos(D3DVECTOR(0,2,0));

	// Start with only only particle system
	m_ParticleSystems[0].ToggleEmitter();

	// Raise the camera
	D3DXMATRIX Mtx, R;
	D3DXMatrixTranslation(&Mtx,0,-3,3);
	D3DXMatrixRotationX(&R, -0.57f);
	Mtx = Mtx*R;
	m_pD3DDev->SetTransform(D3DTRANSFORMSTATE_VIEW, (D3DMATRIX*)&Mtx);

	// Setup the projection matrix
	D3DXMatrixPerspectiveLH(&Mtx, 0.4f, 0.4f*0.75f, 0.2f, 100);
	m_pD3DDev->SetTransform(D3DTRANSFORMSTATE_PROJECTION, (D3DMATRIX*)&Mtx);

	return S_OK;
}

//-----------------------------------------------------------------------------
// Terminate()
//-----------------------------------------------------------------------------
HRESULT CParticleApp::Terminate()
{
	if( m_pParticleTexture )
	{
		m_pParticleTexture->Release();
		m_pParticleTexture = 0;
	}

	m_Renderer.Terminate();

	return S_OK;
}

//-----------------------------------------------------------------------------
// UpdateScene()
//-----------------------------------------------------------------------------
HRESULT CParticleApp::UpdateScene()
{
	// Move the emitters in a circle motion
	D3DVECTOR p;
	p.x =   (float)sin(2.78f*m_SceneTimer.GetFrameTime());
	p.y = 2+(float)cos(2.97f*m_SceneTimer.GetFrameTime());
	p.z =   (float)cos(3.52f*m_SceneTimer.GetFrameTime());
	m_ParticleSystems[0].MoveEmitterPos(p);

	p.y = 2+(float)sin(1.27f*m_SceneTimer.GetFrameTime());
	p.x =   (float)cos(2.37f*m_SceneTimer.GetFrameTime());
	p.z =   (float)cos(3.01f*m_SceneTimer.GetFrameTime());
	m_ParticleSystems[1].MoveEmitterPos(p);

	p.y = 2+(float)sin(3.02f*m_SceneTimer.GetFrameTime());
	p.x =   (float)cos(1.41f*m_SceneTimer.GetFrameTime());
	p.z =   (float)cos(2.89f*m_SceneTimer.GetFrameTime());
	m_ParticleSystems[2].MoveEmitterPos(p);

	// Update particles
	for( int n = 0; n < m_nNumSystems; n++ )
	{
		m_ParticleSystems[n].Update(m_SceneTimer.GetDeltaTime());
	}

	return S_OK;
}

//-----------------------------------------------------------------------------
// RenderScene()
//-----------------------------------------------------------------------------
HRESULT CParticleApp::RenderScene()
{
	// Begin a new scene
    HRESULT hr = m_pD3DDev->BeginScene();
	if( SUCCEEDED( hr ) )
	{
		// Clear the buffers
		m_pD3DX->Clear(D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER);

		// Render the particles
		for( int n = 0; n < m_nNumSystems; n++ )
		{
			m_ParticleSystems[n].Render(&m_Renderer);
		}

		// The scene is finished
        hr = m_pD3DDev->EndScene();
	}

	if( FAILED(hr) )
	{
		ReportError(hr, __FILE__, __LINE__);
		return hr;
	}

	return S_OK;
}


//-----------------------------------------------------------------------------
// MsgProc()
//-----------------------------------------------------------------------------
LRESULT CParticleApp::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if( uMsg == WM_CHAR )
	{
		switch( wParam )
		{
		case '1':
		case '2':
		case '3':
			m_ParticleSystems[wParam - '1'].ToggleEmitter();
			break;
		}
	}

	// Let the superclass take care of the rest of the messages
	return CD3DApp::MsgProc(hWnd, uMsg, wParam, lParam);
}

//-----------------------------------------------------------------------------
// ShowStats()
//-----------------------------------------------------------------------------
HRESULT CParticleApp::ShowStats()
{
    char sz[50];

	// Count the number of live particles
	int n = m_ParticleSystems[0].GetNbrOfLiveParticles();
	n += m_ParticleSystems[1].GetNbrOfLiveParticles();
	n += m_ParticleSystems[2].GetNbrOfLiveParticles();

    sprintf(sz, "%d fps\n%d particles", (int)m_AppTimer.GetFps(), n);
    m_pD3DX->DrawDebugText(0.0f, 0.0f, 0xFFFFFFFF, sz);

	return S_OK;
}