//-----------------------------------------------------------------------------
// Copyright (c) 2000 Andreas Jnsson
//-----------------------------------------------------------------------------
#include "particlesystem.h"

//-----------------------------------------------------------------------------
// CParticleSystem()
//-----------------------------------------------------------------------------
CParticleSystem::CParticleSystem()
{
	// Setup some system attributes
	m_fEnergyDropOff  = 0.3f; 
	m_fBirthInterval  = 0.002f; 
	m_fRadius         = 0.1f;
	m_fGravity        = 0.8f; 
	m_fBirthVelocity  = 0.5f; 
	m_fBounce         = 0.1f;
	m_nLiveParticles  = 0;
	m_fNextBirth      = 0;

	m_nMaxParticles   = 2000;
	m_nColorTableSize = 64;

	m_pParticlePool   = 0;
	m_pFreeParticles  = 0;
	m_pLiveParticles  = 0;

	m_bIsEmitting = false; 
}

//-----------------------------------------------------------------------------
// Initialize()
//-----------------------------------------------------------------------------
HRESULT CParticleSystem::Initialize(D3DXCOLOR c)
{
	m_Color = c;

	// Create the particles and build the linked list
	m_pParticlePool = new SParticle[m_nMaxParticles];
	if( !m_pParticlePool )
		return E_OUTOFMEMORY;

	for( int n = 0; n < m_nMaxParticles-1; n++ )
	{
		m_pParticlePool[n].pNext = &m_pParticlePool[n+1];
	}

	m_pParticlePool[m_nMaxParticles-1].pNext = 0;
	m_pFreeParticles = &m_pParticlePool[0];
	m_pLiveParticles = 0;

	// Create the color lookup table
	m_pColorTable = new D3DCOLOR[m_nColorTableSize];
	if( !m_pColorTable )
		return E_OUTOFMEMORY;

	for( n = 0; n < m_nColorTableSize; n++ )
	{
		D3DXCOLOR LerpColor;
		D3DXColorLerp(&LerpColor, &D3DXCOLOR(0,0,0,0), 
			&m_Color, (float)n / (m_nColorTableSize-1));

		m_pColorTable[n] = (DWORD)LerpColor;
	}

	return S_OK;
}

//-----------------------------------------------------------------------------
// ~CParticleSystem()
//-----------------------------------------------------------------------------
CParticleSystem::~CParticleSystem()
{
	if( m_pParticlePool )
		delete[] m_pParticlePool;

	if( m_pColorTable )
		delete[] m_pColorTable;
}

//-----------------------------------------------------------------------------
// SetEmitterPos()
//-----------------------------------------------------------------------------
void CParticleSystem::SetEmitterPos(D3DVECTOR &Pos)
{
	m_EmitterPos       = Pos;
	m_TargetEmitterPos = Pos;
}

//-----------------------------------------------------------------------------
// MoveEmitterPos()
//-----------------------------------------------------------------------------
void CParticleSystem::MoveEmitterPos(D3DVECTOR &Pos)
{
	m_TargetEmitterPos = Pos;
}

//-----------------------------------------------------------------------------
// Update()
//-----------------------------------------------------------------------------
void CParticleSystem::Update(float fDeltaTime)
{
	m_nLiveParticles = 0;

	// Move and kill particles
	SParticle *pPrev = 0, *pThis = m_pLiveParticles;
	while( pThis )
	{
		// Update the states of the particle
		pThis->Pos     += fDeltaTime * pThis->Vel;
		pThis->fEnergy -= fDeltaTime * m_fEnergyDropOff;
		
		if( pThis->Pos.y > m_fRadius )
			pThis->Vel.y  -= m_fGravity*fDeltaTime;

		// Collision detection against the ground
		if( pThis->Pos.y < m_fRadius )
		{
			// Bounce on the ground
			pThis->Pos.y = m_fRadius;
			pThis->Vel.y = -pThis->Vel.y;
			
			// Retain only some speed after bounce
			pThis->Vel *= m_fBounce;
		}

		// Kill old particles
		if( pThis->fEnergy <= 0 )
		{
			SParticle *pNext = pThis->pNext;

			// Put the dead particle in the list of free particles
			pThis->pNext = m_pFreeParticles;
			m_pFreeParticles = pThis;

			// Reconnect list
			if( pPrev )	
				pPrev->pNext = pNext;

			if( pThis == m_pLiveParticles ) 
				m_pLiveParticles = pNext;

			pThis = pNext;
		}
		else
		{
			m_nLiveParticles++;

			pPrev = pThis;
			pThis = pThis->pNext;
		}
	}

	// Emit new particles
	if( m_bIsEmitting )
	{
		D3DVECTOR EmitterSpeed = m_TargetEmitterPos - m_EmitterPos;

		float fTimeLeft = fDeltaTime;
		while( fTimeLeft > 0 )
		{
			if( fTimeLeft > m_fNextBirth )
			{
				// Move emitter 
				m_EmitterPos += EmitterSpeed * m_fNextBirth / fDeltaTime;

				fTimeLeft -= m_fNextBirth;
				m_fNextBirth = m_fBirthInterval;

				if( m_pFreeParticles )
				{
					// Grab a particle from the list of free particles
					SParticle *pNew = m_pFreeParticles;
					m_pFreeParticles = m_pFreeParticles->pNext;

					// Initialize it
					pNew->Pos     = m_EmitterPos;
					pNew->fEnergy = 1;

					// Initial speed
					float fAlpha = 2 * 3.141592f * rand()/RAND_MAX;
					float fBeta  = 3.141592f * ((float)rand()/RAND_MAX - 0.5f); 
					pNew->Vel.x  = m_fBirthVelocity*(float)(cos(fBeta) * sin(fAlpha));
					pNew->Vel.y  = m_fBirthVelocity*(float)sin(fBeta);
					pNew->Vel.z  = m_fBirthVelocity*(float)(cos(fBeta) * cos(fAlpha));

					// Insert the particle into the list of live particles
					pNew->pNext = m_pLiveParticles;
					m_pLiveParticles = pNew;

					m_nLiveParticles++;
				}
			}
			else
			{
				// Move emitter 
				m_EmitterPos = m_TargetEmitterPos;

				m_fNextBirth -= fTimeLeft;
				fTimeLeft = 0;
			}
		}
	}
	else
	{
		m_EmitterPos = m_TargetEmitterPos;
	}
}

//-----------------------------------------------------------------------------
// Render()
//-----------------------------------------------------------------------------
HRESULT CParticleSystem::Render(CRenderer *pRenderer)
{
	HRESULT hr;

	// Get the view matrix
	D3DMATRIX Mtx;
	pRenderer->m_pD3DDev->GetTransform(D3DTRANSFORMSTATE_VIEW, &Mtx);

	// Extract the camera's up and right vector
	D3DXVECTOR3 cy, cx;
	cx.x = Mtx._11; cx.y = Mtx._21; cx.z = Mtx._31; // Right
	cy.x = Mtx._12; cy.y = Mtx._22; cy.z = Mtx._32; // Up

	// Setup the particle vertices so that the face of  
	// the particles are always directed at the camera
	D3DXVECTOR3 v1,v2,v3;
	v1 = m_fRadius*cy*2;
	v2 = m_fRadius*( cx*1.732f - cy);
	v3 = m_fRadius*(-cx*1.732f - cy);

	// Render the particles
	if( FAILED( hr = pRenderer->Begin() ) )
		return hr;

	SParticle *p = m_pLiveParticles;
	while( p )
	{
		// Fade particle color
		D3DCOLOR c = m_pColorTable[(int)(p->fEnergy*m_nColorTableSize)];

		// Build particle vertices
		pRenderer->Vertex(p->Pos + v1, c, D3DXVECTOR2(0.5f,0.134f));
		pRenderer->Vertex(p->Pos + v2, c, D3DXVECTOR2(1   ,1     ));
		pRenderer->Vertex(p->Pos + v3, c, D3DXVECTOR2(0   ,1     ));

		p = p->pNext;
	}

	if( FAILED( hr = pRenderer->End() ) )
		return hr;

	return S_OK;
}


//-----------------------------------------------------------------------------
// ToggleEmitter()
//-----------------------------------------------------------------------------
void CParticleSystem::ToggleEmitter()
{
	m_bIsEmitting = !m_bIsEmitting;
	m_fNextBirth  = 0;
}

//-----------------------------------------------------------------------------
// GetNbrOfLiveParticles()
//-----------------------------------------------------------------------------
int CParticleSystem::GetNbrOfLiveParticles()
{
	return m_nLiveParticles;
}