//-----------------------------------------------------------------------------
// Copyright (c) 2000 Andreas Jnsson
//-----------------------------------------------------------------------------
#include "renderer.h"

//-----------------------------------------------------------------------------
// CRenderer()
//-----------------------------------------------------------------------------
CRenderer::CRenderer()
{
	m_nBufferSize   = 10000;
	m_pD3DDev       = 0;
	m_pVertexBuffer = 0;
	m_pwIndices     = 0;
	m_pVertices     = 0;
	m_nVertexOffset = 0;
}

//-----------------------------------------------------------------------------
// Initialize()
//-----------------------------------------------------------------------------
HRESULT CRenderer::Initialize(ID3DXContext *pD3DX, int nMaxTriangles)
{
	HRESULT hr;
	
	// Store the D3D objects for later use
	m_pD3DX = pD3DX;
	m_pD3DX->AddRef();

	m_pD3D    = m_pD3DX->GetD3D();
	m_pD3DDev = m_pD3DX->GetD3DDevice();

	// Create a vertex buffer
	m_nMaxBatchSize = nMaxTriangles*3;

	D3DVERTEXBUFFERDESC VBDesc;
	VBDesc.dwSize        = sizeof(D3DVERTEXBUFFERDESC);
	VBDesc.dwCaps        = D3DVBCAPS_WRITEONLY; // We don't want to read 
												// from the buffer
	VBDesc.dwFVF         = FVF_VERTEX;
	VBDesc.dwNumVertices = m_nBufferSize;

	// Because our vertex buffer will be updated every frame we will
	// put it in system memory so that the whole vertex buffer is transferred
	// to the card in one large batch instead of each vertex by themselves.
	VBDesc.dwCaps |= D3DVBCAPS_SYSTEMMEMORY;

	if(FAILED(hr = m_pD3D->CreateVertexBuffer(&VBDesc, &m_pVertexBuffer, 0)))
	    return hr;

	return S_OK;
}

//-----------------------------------------------------------------------------
// Terminate()
//-----------------------------------------------------------------------------
HRESULT CRenderer::Terminate()
{
	if( m_pVertexBuffer )
	{
		m_pVertexBuffer->Release();
		m_pVertexBuffer = 0;
	}

	if( m_pD3D )
	{
		m_pD3D->Release();
		m_pD3D = 0;
	}

	if( m_pD3DDev )
	{
		m_pD3DDev->Release();
		m_pD3DDev = 0;
	}

	if( m_pD3DX )
	{
		m_pD3DX->Release();
		m_pD3DX = 0;
	}

	return S_OK;
}

//-----------------------------------------------------------------------------
// Begin()
//-----------------------------------------------------------------------------
HRESULT CRenderer::Begin()
{
	HRESULT hr;

	// Start on a new list of vertices
	m_nNumVertices = 0;

	// Get a pointer to our vertices
	if( m_nVertexOffset + m_nMaxBatchSize > m_nBufferSize )
	{
		// Because the graphic accelerator may still be 
		// using the old vertices we tell Direct3D that it 
		// can give us some other memory, we wont use the old
		hr = m_pVertexBuffer->Lock(DDLOCK_WAIT | DDLOCK_WRITEONLY |	   
			DDLOCK_DISCARDCONTENTS, (void **)&m_pVertices, NULL);

		m_nVertexOffset = 0;
	}
	else
	{
		// Again, the graphic accelerator may still be
		// using the old vertices but this time we have more
		// room in the buffer so we only tell Direct3D that we
		// will not overwrite previous vertices so it doesn't
		// need to interrupt ongoing processing
		hr = m_pVertexBuffer->Lock(DDLOCK_WAIT | DDLOCK_WRITEONLY |
			DDLOCK_NOOVERWRITE, (void **)&m_pVertices, NULL);

		// Skip the old vertices so we don't overwrite them
		m_pVertices += m_nVertexOffset;
	}


	if( FAILED(hr) )
		return hr;

	return S_OK;
}

//-----------------------------------------------------------------------------
// Vertex()
//-----------------------------------------------------------------------------
void CRenderer::Vertex(D3DXVECTOR3 &v, D3DCOLOR &c, D3DXVECTOR2 &t)
{
	if( m_pVertices == 0 )
		return;

	m_pVertices->v = v;
	m_pVertices->c = c;
	m_pVertices->t = t;
	m_pVertices++;

	m_nNumVertices++;

	if( m_nNumVertices == m_nMaxBatchSize )
	{
		// Call end to make sure the contents is rendered
		if( FAILED(End()) )
			return;

		// Start on a clean vertex list
		if( FAILED(Begin()) )
			return;
	}
}

//-----------------------------------------------------------------------------
// End()
//-----------------------------------------------------------------------------
HRESULT CRenderer::End()
{
	m_pVertexBuffer->Unlock();

	// Render the triangles
	if( m_nNumVertices )
	{
		HRESULT hr;
		hr = m_pD3DDev->DrawPrimitiveVB(D3DPT_TRIANGLELIST,
			m_pVertexBuffer, m_nVertexOffset, m_nNumVertices, 0);

		m_nVertexOffset += m_nNumVertices;

		return hr;
	}

	return S_OK;
}